@@ -1731,7 +1731,6 @@ func_exit:
 void
 btr_search_update_hash_on_insert(btr_cur_t* cursor)
 {
-	hash_table_t*	table;
 	buf_block_t*	block;
 	dict_index_t*	index;
 	const rec_t*	rec;
@@ -1798,6 +1797,8 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor)
 	}
 
 	rw_lock_t* const latch = btr_get_search_latch(index);
+	/* We must not look up "table" before acquiring the latch. */
+	hash_table_t* table = NULL;
 	bool locked = false;
 
 	if (!page_rec_is_infimum(rec)) {
@@ -1830,6 +1831,7 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor)
 			if (!btr_search_enabled || !block->index) {
 				goto function_exit;
 			}
+
 			table = btr_get_search_table(index);
 		}
 
@@ -1851,6 +1853,7 @@ check_next_rec:
 				if (!btr_search_enabled || !block->index) {
 					goto function_exit;
 				}
+
 				table = btr_get_search_table(index);
 			}
 
@@ -1868,6 +1871,7 @@ check_next_rec:
 			if (!btr_search_enabled || !block->index) {
 				goto function_exit;
 			}
+
 			table = btr_get_search_table(index);
 		}
 
