@@ -445,9 +445,11 @@ row_vers_must_preserve_del_marked(
 @param[in]	clust_index	clustered index
 @param[in]	index		the secondary index
 @param[in]	heap		heap used to build virtual dtuple
-@param[in,out]	vcol_info	virtual column information. */
+@param[in,out]	vcol_info	virtual column information.
+@return		true in case of success
+		false if virtual column computation fails */
 static
-void
+bool
 row_vers_build_clust_v_col(
 	dtuple_t*		row,
 	dict_index_t*		clust_index,
@@ -455,11 +457,8 @@ row_vers_build_clust_v_col(
 	mem_heap_t*		heap,
 	purge_vcol_info_t*	vcol_info)
 {
-	mem_heap_t*	local_heap = NULL;
-	VCOL_STORAGE	*vcol_storage= NULL;
 	THD*		thd= current_thd;
 	TABLE*		maria_table= 0;
-	byte*		record= 0;
 
 	ut_ad(dict_index_has_virtual(index));
 	ut_ad(index->table == clust_index->table);
@@ -469,15 +468,13 @@ row_vers_build_clust_v_col(
 		maria_table = vcol_info->table();
 	}
 
-	innobase_allocate_row_for_vcol(thd, index,
-				       &local_heap,
-				       &maria_table,
-				       &record,
-				       &vcol_storage);
+	ib_vcol_row vc(NULL);
+	byte *record = vc.record(thd, index, &maria_table);
 
 	if (vcol_info && !vcol_info->table()) {
 		vcol_info->set_table(maria_table);
-		goto func_exit;
+		// wait for second fetch
+		return true;
 	}
 
 	for (ulint i = 0; i < dict_index_get_n_fields(index); i++) {
@@ -490,19 +487,18 @@ row_vers_build_clust_v_col(
 			col = reinterpret_cast<const dict_v_col_t*>(
 				ind_field->col);
 
-			innobase_get_computed_value(
-				row, col, clust_index, &local_heap,
+			dfield_t *vfield = innobase_get_computed_value(
+				row, col, clust_index, &vc.heap,
 				heap, NULL, thd, maria_table, record, NULL,
 				NULL, NULL);
+			if (vfield == NULL) {
+				innobase_report_computed_value_failed(row);
+				ut_ad(0);
+				return false;
+			}
 		}
 	}
-
-func_exit:
-	if (local_heap) {
-		if (vcol_storage)
-			innobase_free_row_for_vcol(vcol_storage);
-		mem_heap_free(local_heap);
-	}
+	return true;
 }
 
 /** Build latest virtual column data from undo log
@@ -839,8 +835,11 @@ row_vers_build_cur_vrow(
 			mtr->commit();
 		}
 
-		row_vers_build_clust_v_col(
+		bool res = row_vers_build_clust_v_col(
 			row, clust_index, index, heap, vcol_info);
+		if (!res) {
+			return NULL;
+		}
 
 		if (vcol_info != NULL && vcol_info->is_first_fetch()) {
 			return NULL;
@@ -961,10 +960,14 @@ row_vers_old_has_index_entry(
 					mtr->commit();
 				}
 
-				row_vers_build_clust_v_col(
+				bool res = row_vers_build_clust_v_col(
 					row, clust_index, index, heap,
 					vcol_info);
 
+				if (!res) {
+					goto unsafe_to_purge;
+				}
+
 				if (vcol_info && vcol_info->is_first_fetch()) {
 					goto unsafe_to_purge;
 				}
